import { filter, interval, map, Observable, Observer, Subject } from "rxjs";
import { Bus, FisMessage, MessageTransmissionSet } from "../interface/transport.interface";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { TransportEvent, TransportMessage } from "../interface/connector.interface";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";

class Supervisor {
    private generalBus: Subject<TransportEvent> = new Subject()
    private console = new ConsoleLogger('Supervisor', ['base'])
    private isClient: boolean = true
    private transmissionManager!: MessageTransmissionManager
    private event: Observable<TransportEvent>
    private transmissionSets: MessageTransmissionSet[] = []
    private outgoingPipe: Subject<any> = new Subject()

    constructor() {
        this.transmissionManager = new MessageTransmissionManager(this.isClient)
        this.event = this.transmissionManager.getEvent()

        this.transmissionManager.subscribe().subscribe((transmissionSet: MessageTransmissionSet) => {
            this.console.log({ message: `Acquired transmission set for client` })
            this.transmissionSets.push(transmissionSet)

            this.handleActivity(transmissionSet)
            this.outgoingPipe.subscribe(message => transmissionSet.transmitter.emit(message))
        })
    }

    // only called once for each connected clients.
    private handleActivity(messageTransmission: MessageTransmissionSet): void {
        // start listening to incoming messages from this client
        messageTransmission.receiver.getMessageBus(Bus.GeneralBus).subscribe((event: TransportEvent) => {
            this.console.log({ message: `General Bus ${event.event}`, details: event })
            this.generalBus.next(event)
        })

        let request: FisMessage = {
            header: {
                messageID: uuidv4(),
                messageName: 'RequestMessage'
            },
            data: 'Data'
        }

        // this.request(request, messageTransmission).subscribe({
        //     next: res => this.console.log({ message: `received ${res.header.messageID}`, details: res }),
        //     complete: () => this.console.log({ message: `Responses Completed for request: ${request.header.messageID}` })
        // })

        // this.startGeneratingRequest(1000, this.outgoingPipe)
    }

    private request(request: FisMessage, messageTransmission: MessageTransmissionSet): Observable<any> {
        return new Observable((response: Observer<any>) => {
            messageTransmission.transmitter.emit(request)
            this.generalBus.pipe(
                filter(event => event.event == 'New Message'),
                filter(event => (((event.data as TransportMessage)?.payload as WrappedMessage)?.payload as FisMessage)?.header.messageID === request.header.messageID),
                map(event => (((event.data as TransportMessage)?.payload as WrappedMessage)?.payload as FisMessage))
            ).subscribe(message => {
                if (message.data == 'Complete') {
                    response.complete()
                } else {
                    response.next(message)
                }
            })
        })
    }


    private startGeneratingRequest(intervalDuration: number, requestsPipe: Subject<FisMessage>) {
        interval(intervalDuration).subscribe(time => {
            let message: FisMessage = {
                header: {
                    messageID: uuidv4(),
                    messageName: 'NotificationMessage'
                },
                data: 'Data'
            }
            requestsPipe.next(message)
        })
    }

}

let supervisor = new Supervisor()
