import { filter, interval, map, Observable, Observer, Subject, take } from "rxjs";
import { Bus, EventMessage, FisMessage, MessageTransmissionSet } from "../interface/transport.interface";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { TransportEvent, TransportMessage } from "../interface/connector.interface";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
class Supervisor {
    private console = new ConsoleLogger('Supervisor', ['base'])
    private clientIncomingMessage: Subject<FisMessage> = new Subject()
    private messageProducer!: MessageProducer
    private transmissionManager!: MessageTransmissionManager
    private event!: Observable<TransportEvent>
    private transmissionSets: MessageTransmissionSet[] = []

    constructor() {
        // so need them adapters now. But supervisor shouldn't be concerned, only messageTransmissionManager and ConnectionManager
        this.messageProducer = new MessageProducer(this.clientIncomingMessage)
        this.transmissionManager = new MessageTransmissionManager()
        this.event = this.transmissionManager.getEvent()

        this.transmissionManager.subscribe().subscribe((transmissionSet: MessageTransmissionSet) => {
            this.transmissionSets.push(transmissionSet)

            this.handleClientActivity(transmissionSet)
        })
    }

    // only called once for each connected clients.
    private handleClientActivity(messageTransmission: MessageTransmissionSet): void {
        // start listening to incoming messages from this client
        messageTransmission.receiver.getMessageBus(Bus.GeneralBus).subscribe((event: TransportEvent) => {
            this.console.log({ message: `General Bus`, details: event }) // receiving end
            let requestMessage: FisMessage = ((event.data as TransportMessage).payload as WrappedMessage).payload as FisMessage
            // this.clientIncomingMessage.next(requestMessage)
            this.messageProducer.getOutgoingMessages().pipe(
                filter(message => message.header.messageID === requestMessage.header.messageID)
            ).subscribe(message => {
                messageTransmission.transmitter.emit(message)
            })
        })

        // to emulate general notification. Send every second
        // this.messageProducer.getNotificationMessage().subscribe((message: FisMessage) => {
        //     messageTransmission.transmitter.emit(message)
        // })
    }
}


class MessageProducer {
    private console = new ConsoleLogger('Message Producer', ['base'])
    private generalNotification: Subject<FisMessage> = new Subject()
    private incomingMessageBus!: Subject<FisMessage>
    private outgoingMessageBus: Subject<FisMessage> = new Subject()

    constructor(incomingMessageBus: Subject<FisMessage>) {
        this.console.log({ message: `Constructing Message Producer`})
        this.incomingMessageBus = incomingMessageBus

        this.generateNotifcation().subscribe(this.generalNotification)
        this.handleIncomingRequests(this.incomingMessageBus.asObservable(), this.outgoingMessageBus)
    }

    public getNotificationMessage(): Observable<FisMessage> {
        return this.generalNotification.asObservable()
    }

    public getOutgoingMessages(): Observable<FisMessage> {
        return this.outgoingMessageBus.asObservable()
    }

    // this is called no problem
    private handleIncomingRequests(requests: Observable<FisMessage>, outgoingMessageBus: Subject<FisMessage>): void {
        requests.subscribe((request: FisMessage) => {
            this.console.log({ message: `Generating response for new request ${request.header.messageID}`})
            this.generateMessage(request.header.messageID, 10).subscribe({
                next: message => outgoingMessageBus.next(message),
                error: error => this.console.log({ message: 'observer Error', details: error }),
                complete: () => {
                    outgoingMessageBus.next({
                        header: {
                            messageID: request.header.messageID,
                            messageName: `ResponseMessage`
                        },
                        data: `Complete`
                    } as FisMessage)
                }
            })
        })
    }

    private generateMessage(requestID: string, amount: number): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                take(amount), // Ensures only 'amount' messages are generated
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }

    private generateNotifcation(): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }
}

let supervisor = new Supervisor()
